@@ -14,9 +14,7 @@
 
 import getpass
 import re
-from typing import Optional
 
-import attr
 import click
 
 import releasetool.circleci
@@ -24,19 +22,16 @@ import releasetool.git
 import releasetool.github
 import releasetool.secrets
 import releasetool.commands.common
+from releasetool.commands.common import TagContext
+
+# repos that still use kokoro for publication:
+nodejs_handle_publish = [
+    "googleapis/google-api-nodejs-client",
+    "googleapis/nodejs-logging",
+]
 
 
-@attr.s(auto_attribs=True, slots=True)
-class Context(releasetool.commands.common.GitHubContext):
-    package_name: Optional[str] = None
-    release_pr: Optional[dict] = None
-    release_tag: Optional[str] = None
-    release_version: Optional[str] = None
-    release_notes: Optional[str] = None
-    github_release: Optional[dict] = None
-
-
-def determine_release_pr(ctx: Context) -> None:
+def determine_release_pr(ctx: TagContext) -> None:
     click.secho(
         "> Let's figure out which pull request corresponds to your release.", fg="cyan"
     )
@@ -53,7 +48,7 @@ def determine_release_pr(ctx: Context) -> None:
     ctx.release_pr = pulls[pull_idx - 1]
 
 
-def determine_release_tag(ctx: Context) -> None:
+def determine_release_tag(ctx: TagContext) -> None:
     click.secho("> Determining what the release tag should be.", fg="cyan")
     head_ref = ctx.release_pr["head"]["ref"]
     match = re.match("release-(.+)", head_ref)
@@ -73,23 +68,39 @@ def determine_release_tag(ctx: Context) -> None:
     click.secho(f"Release tag is {ctx.release_tag}.")
 
 
-def determine_package_version(ctx: Context) -> None:
+def determine_package_version(ctx: TagContext) -> None:
     click.secho("> Determining the package version.", fg="cyan")
-    match = re.match("(?P<version>v?\d+?\.\d+?\.\d+?)", ctx.release_tag)
+    match = re.match(r"(?P<version>v?\d+?\.\d+?\.\d+?)", ctx.release_tag)
     ctx.release_version = match.group("version")
-    click.secho(
-        f"package version: {ctx.release_version}."
-    )
+    click.secho(f"package version: {ctx.release_version}.")
 
 
-def get_release_notes(ctx: Context) -> None:
+def determine_kokoro_job_name(ctx: TagContext) -> None:
+    if ctx.upstream_repo in nodejs_handle_publish:
+        ctx.kokoro_job_name = (
+            f"cloud-devrel/client-libraries/nodejs/release/{ctx.upstream_repo}/publish"
+        )
+    else:
+        ctx.kokoro_job_name = (
+            f"cloud-devrel/client-libraries/nodejs/release/{ctx.upstream_repo}/docs"
+        )
+
+
+def get_release_notes(ctx: TagContext) -> None:
     click.secho("> Grabbing the release notes.")
     changelog = ctx.github.get_contents(
         ctx.upstream_repo, "CHANGELOG.md", ref=ctx.release_pr["merge_commit_sha"]
     ).decode("utf-8")
 
+    _get_latest_release_notes(ctx, changelog)
+
+
+def _get_latest_release_notes(ctx: TagContext, changelog: str):
+    # the 'v' prefix is not used in the conventional-changelog templates
+    # used in automated CHANGELOG generation:
+    version = re.sub(r"^v", "", ctx.release_version)
     match = re.search(
-        rf"## {ctx.release_version}\n(?P<notes>.+?)(\n##\s|\Z)",
+        rf"## v?\[?{version}[^\n]*\n(?P<notes>.+?)(\n##\s|\n### \[?[0-9]+\.|\Z)",
         changelog,
         re.DOTALL | re.MULTILINE,
     )
@@ -99,27 +110,31 @@ def get_release_notes(ctx: Context) -> None:
         ctx.release_notes = ""
 
 
-def create_release(ctx: Context) -> None:
+def create_release(ctx: TagContext) -> None:
     click.secho("> Creating the release.")
 
     ctx.github_release = ctx.github.create_release(
         repository=ctx.upstream_repo,
         tag_name=ctx.release_version,
-        target_committish=ctx.release_pr["merge_commit_sha"],
+        target_commitish=ctx.release_pr["merge_commit_sha"],
         name=f"{ctx.release_version}",
         body=ctx.release_notes,
     )
 
     release_location_string = f"Release is at {ctx.github_release['html_url']}"
     click.secho(release_location_string)
-    click.secho("CI will handle publishing the package to PyPI.")
+    click.secho("CI will handle publishing the package to npm.")
 
     ctx.github.create_pull_request_comment(
         ctx.upstream_repo, ctx.release_pr["number"], release_location_string
     )
 
+    ctx.github.update_pull_labels(
+        ctx.release_pr, add=["autorelease: tagged"], remove=["autorelease: pending"]
+    )
 
-def wait_on_circle(ctx: Context) -> None:
+
+def wait_on_circle(ctx: TagContext) -> None:
     circle = releasetool.circleci.CircleCI(repository=ctx.upstream_repo)
     click.secho("> Waiting for CircleCI to queue a release build")
     tag_name = ctx.release_version
@@ -134,19 +149,34 @@ def wait_on_circle(ctx: Context) -> None:
         click.secho(f"CircleCI Build not found for tag {tag_name}...")
 
 
-def tag() -> None:
-    ctx = Context()
+def tag(ctx: TagContext = None) -> TagContext:
+    if not ctx:
+        ctx = TagContext()
 
-    click.secho(f"o/ Hey, {getpass.getuser()}, let's tag a release!", fg="magenta")
+    if ctx.interactive:
+        click.secho(f"o/ Hey, {getpass.getuser()}, let's tag a release!", fg="magenta")
 
-    releasetool.commands.common.setup_github_context(ctx)
+    if ctx.github is None:
+        releasetool.commands.common.setup_github_context(ctx)
+
+    if ctx.release_pr is None:
+        determine_release_pr(ctx)
 
-    determine_release_pr(ctx)
     determine_release_tag(ctx)
     determine_package_version(ctx)
+
+    # If the release already exists, don't do anything
+    if releasetool.commands.common.release_exists(ctx):
+        click.secho(f"{ctx.release_tag} already exists.", fg="magenta")
+        return ctx
+
     get_release_notes(ctx)
-
     create_release(ctx)
-    wait_on_circle(ctx)
 
-    click.secho(f"\o/ All done!", fg="magenta")
+    determine_kokoro_job_name(ctx)
+    releasetool.commands.common.publish_via_kokoro(ctx)
+
+    if ctx.interactive:
+        click.secho("\\o/ All done!", fg="magenta")
+
+    return ctx
